
/**
 * @file Astri_MA_Processor.h
 * @brief This file contains the declaration of the AstriMaProcessor class and AstriMaProcBuilder class.
 * 
 * The AstriMaProcessor class is a derived class of BaseDAQ and represents the processor for ASTRI packets.
 * It provides methods for starting and stopping the processor, as well as configuring the processor using the AstriMaProcBuilder class.
 * 
 * The AstriMaProcBuilder class is a builder class for configuring the AstriMaProcessor.
 * It provides methods for setting the receiver, provider, and packets to process, as well as other configuration options.
 */

#include <Base_DAQ.h>
#include <map>
#include <Packet_Structure_Json.h>
#include <Base_Configurator.h>
#include <Astri_MA_Generic.h>
#include <thread>

#ifndef ASTRI_PACKETS_PATH
#define ASTRI_PACKETS_PATH std::string(std::getenv("HOME")).append("/BIAS/config/AstriPackets")
#endif

namespace inaf::oasbo::Processors {

class AstriMaProcessor: public inaf::oasbo::DAQ::BaseDAQ {
protected:
	std::map<std::string, Packets::PacketStructureJson*> packetStructuresMap; /**< Map of the structures of the packets.
	 * The key is the packet key, i.e. "22" for the S22 pakcets etc.
	 * The value is a pointer to the corresponding PacketStructureJson object.
	 */
	std::map<std::string, bool> packetsFlags; /**< Map of packets to process. 
	 * the key is the packet key, i.e. "22" for the S22 pakcets etc.
	 * the value is a boolean flag, true if the packet is to be processed, false otherwise.
	 */
	void populatePacketStructuresMap(); /**< Method for generating the packet structures map. */
	int closingTimeout; /**< Timeout for stopping the processor, i.e. for the cleanup process. */
	int receiveAndProcessPacket(); /**< Method for receiving and processing a packet. */
	std::string addPrefixToKey(std::string packetKey); /**< Helper method for adding a prefix to a packet key (i.e the "S" to 22 etc). */
	int connectReceiver(); /**< Method for connecting the receiver. */
	int connectProvider(); /**< Method for connecting the provider. */
	int connectionTimeout; /**< Timeout for establishing the connection. */
	void printLog(std::ostream &os, std::string message); /**< Method for printing log messages. */
	void cleanup(); /**< Method for cleaning up resources. */
	std::thread receiveAndProcessThread; /**< Thread for receiving and processing packets. */
	void receiveAndProcessThreadFunction(); /** Function for the receiveAndProcess Thread. */
	bool receiveAndProcessFlag = false; /**< Flag for stopping the receiving ang processing */

public:
	AstriMaProcessor(); /**< Constructor for AstriMaProcessor. */
	~AstriMaProcessor(); /**< Destructor for AstriMaProcessor. */
	void start() override; /**< Method for starting the processor. */
	void stop() override; /**< Method for stopping the processor. */
	void stop(int seconds); /**< Method for stopping the processor after a specified number of seconds. */
	void switchState(const Status) override; /**< Method for switching the state of the processor. */
	int deliverPacket() override; /**< Method for delivering a packet. */
	std::string getStateStr(Status) override; /**< Method for getting the state as a string. */
	friend class AstriMaProcBuilder;
	/**< Friend class declaration for AstriMaProcBuilder. */
};

class AstriMaProcBuilder {
protected:
	AstriMaProcessor *proc; /**< Pointer to the AstriMaProcessor being built. */
public:
	AstriMaProcBuilder(); /**< Constructor for AstriMaProcBuilder. */
	~AstriMaProcBuilder(); /**< Destructor for AstriMaProcBuilder. */
	void reset(); /**< Method for resetting the builder. */
	AstriMaProcBuilder* configFrom(Configurators::BaseConfigurator &conf); /**< Method for configuring the builder from a BaseConfigurator. */
	AstriMaProcBuilder* setReceiver(Receivers::BaseReceiver*); /**< Method for setting the receiver. */
	AstriMaProcBuilder* setProvider(Providers::BaseProvider*); /**< Method for setting the provider. */
	AstriMaProcBuilder* setPacket(Packets::BasePacket*); /**< Method for setting the packet. */
	AstriMaProcBuilder* addPacketToProcess(std::string); /**< Method for adding a packet to process. */
	AstriMaProcBuilder* removePacketToProcess(std::string); /**< Method for removing a packet to process. */
	AstriMaProcessor* getProcessor(); /**< Method for getting the built AstriMaProcessor. */

	static const std::string config_target; /**< Configuration target constant. */
	static const std::string connectionTimeout_key; /**< Connection timeout key constant. */
	static const std::string closingTimeout_key; /**< stopping timeout key constant. */
	static const std::string receiver_key; /**< Receiver key constant. */
	static const std::string provider_key; /**< Provider key constant. */
	static const std::string packet_key; /**< Packet key constant. */
	static const std::string typesubtype_key; /**< Type subtype key constant. */
	static const std::string redis_rcv_key; /**< Redis receiver key constant. */
	static const std::string file_rcv_key; /**< File receiver key constant. */
	static const std::string kafkaavro_prov_key; /**< Kafka Avro provider key constant. */
	static const std::string astri_ma_packet_key; /**< ASTRI MA packet key constant. */
	static const std::string astri_horn_packet_key; /**< ASTRI horn packet key constant. */
};

}
