#include <Astri_MA_Processor.h>
#include <Astri_MA_Generic.h>
#include <Redis_Receiver.h>
#include <KafkaAvro_Provider.h>

#include <ctime>
#include <iomanip>

using namespace inaf::oasbo::Processors;

void AstriMaProcessor::printLog(std::ostream &os, std::string message) {
	time_t now = time(nullptr);
	os << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[Astri MA Processor]" << "\t" << message << std::endl;
}

AstriMaProcessor::AstriMaProcessor() {
	this->receiver = Receivers::RedisReceiverBuilder().getReceiver();
	this->provider = Providers::KafkaAvroProviderBuilder().getProvider();
	std::string ma_packet(
			std::string(ASTRI_PACKETS_PATH).append(
					"/AstriMA/AstriMAGeneric.json"));
	Packets::PacketStructureJson structure(ma_packet);
	this->packet = new Packets::AstriMaGeneric(structure);

	this->closingTimeout = 10;
	this->connectionTimeout = 10;
	this->currentState = Status::INIT;
}

AstriMaProcessor::~AstriMaProcessor() {
	for (auto &pair : packetStructuresMap)
		delete pair.second;
	if (receiveAndProcessThread.joinable()) {
		receiveAndProcessThread.join();
	}
}

void AstriMaProcessor::switchState(Status newState) {
	if (newState == STOP) {
		setCurrentState(Status::STOP);
		this->receiveAndProcessFlag = false;
		if (receiveAndProcessThread.joinable()) {
			receiveAndProcessThread.join();
		}
		return;
	}

	switch (this->currentState) {
	case Status::INIT: {
		switch (newState) {
		case Status::READY: // from INIT to READY
			if (connectReceiver() < 0 || connectProvider() < 0) {
				break;
			}
			this->setCurrentState(Status::READY);
			break;
		default:
			std::stringstream ss;
			ss << "Switch from " << getStateStr(this->currentState) << " to "
					<< getStateStr(newState) << " not allowed";
			printLog(std::cerr, ss.str());
			break;
		}
		break;
	}
	case Status::READY: {
		switch (newState) {
		case Status::RUN: // from READY to RUN
			this->setCurrentState(newState);
			if (!this->receiveAndProcessThread.joinable()) { // Start the receiveAndProcessThread only the first time the state is set to RUN from READY
				this->receiveAndProcessThread = std::thread(&AstriMaProcessor::receiveAndProcessThreadFunction, this);
			}
			break;
		default:
			std::stringstream ss;
			ss << "Switch from " << getStateStr(this->currentState) << " to "
					<< getStateStr(newState) << " not allowed";
			printLog(std::cerr, ss.str());
			break;
		}
		break;
	}
	case Status::RUN: {
		switch (newState) {
		case Status::RUN:
			break;
		case Status::READY: // from RUN to READY
			if (connectReceiver() < 0 || connectProvider() < 0) { // recv or provider connection went down, exit...
				this->receiveAndProcessFlag = false;
				break;
			}
			this->setCurrentState(Status::READY);
			this->switchState(RUN);
			break;
		default:
			break;
		}
		break;
	}
	default:
		break;
	}
}



void AstriMaProcessor::receiveAndProcessThreadFunction() {
	this->receiveAndProcessFlag = true;
	while (this->receiveAndProcessFlag) { // becomes false if stop command or connection timeout exceeded (in next line switchState(Status::READY);)
		int res = receiveAndProcessPacket();
		if (res <= 0 && receiveAndProcessFlag) { // recv or provider connection went down
			switchState(Status::READY);
		}
	}
	this->cleanup();
	this->setCurrentState(STOP);
}

int AstriMaProcessor::connectReceiver() {
	if (receiver->isConnectedToClient()){
		return 1;
	}
	if (receiver->connectToClient() > 0){
		return 1;
	}
	auto startTime = std::chrono::steady_clock::now();
	auto currentTime = std::chrono::steady_clock::now();
	auto elapsedTime = std::chrono::duration_cast<std::chrono::seconds>(
			currentTime - startTime).count();
	while ((this->connectionTimeout < 0 || elapsedTime < this->connectionTimeout)
			&& (this->currentState != Status::STOP)) {
		if (receiver->connectToClient() > 0)
			return 1;
		else { // try every second
			std::this_thread::sleep_for(std::chrono::seconds(1));
			currentTime = std::chrono::steady_clock::now();
			elapsedTime = std::chrono::duration_cast<std::chrono::seconds>(
					currentTime - startTime).count();
		}
	}
	printLog(std::cerr, "Unable to establish connection to receiver");
	return -1;
}

int AstriMaProcessor::connectProvider() {
	if (provider->isOpen())
		return 1;
	if (provider->open() > 0)
		return 1;
	auto startTime = std::chrono::steady_clock::now();
	auto currentTime = std::chrono::steady_clock::now();
	auto elapsedTime = std::chrono::duration_cast<std::chrono::seconds>(
			currentTime - startTime).count();
	while ((this->connectionTimeout < 0 || elapsedTime < this->connectionTimeout)
			&& (this->currentState != Status::STOP)) {
		if (provider->open() > 0)
			return 1;
		else { // try every second
			std::this_thread::sleep_for(std::chrono::seconds(1));
			currentTime = std::chrono::steady_clock::now();
			elapsedTime = std::chrono::duration_cast<std::chrono::seconds>(
					currentTime - startTime).count();
		}
	}
	printLog(std::cerr, "Unable to establish connection to provider");
	return -1;
}

int AstriMaProcessor::deliverPacket() {
	Packets::AstriMaGeneric *packet =
			static_cast<Packets::AstriMaGeneric*>(this->packet); // From basePacket to AstriPacket to use getType and getSubType
	std::string flagKey = std::string(std::to_string(packet->getType())).append(
			std::to_string(packet->getSubType()));
	int res = 1;
	try {
		if (packetsFlags[flagKey]) {
			Packets::BasePacketStructure *previous =
					&packet->getPacketStructure(); // Save the previous packet structure which was the generic one
			packet->updatePacketStructure(*packetStructuresMap[flagKey]); // update tghe packet structure with the specific one (i.e. C11, C14, S22, etc)
			std::string key = addPrefixToKey(flagKey); // UNcomment if you want to dispatch packets based on type_subype
			res = provider->write(*packet,provider->getDest()+"_"+key); // UNcomment if you want to dispatch packets based on type_subype
//			res = provider->write(*packet); // comment if you want to dispatch packets based on type_subype
			packet->updatePacketStructure(*previous); // restore the generic packet structure to be able to receive the next packet
		}
	} catch (std::out_of_range &ex) {
		printLog(std::cerr, "Packet type_subtype not recognized: " + flagKey);
		return -1;
	}
	return res;
}

int AstriMaProcessor::receiveAndProcessPacket() {
	int rcv = receiver->receiveFromClient(*packet);
	if (rcv > 0) {
		return this->deliverPacket();
	} else {
		return rcv;
	}
}

void AstriMaProcessor::stop() {
	stop(this->closingTimeout);
}
void AstriMaProcessor::stop(int seconds) {
	this->closingTimeout = seconds;
	switchState(Status::STOP);
}

void AstriMaProcessor::start() {
	populatePacketStructuresMap();
	switchState(Status::READY);
	if (this->currentState != Status::READY)
		switchState(Status::STOP);
	else {
		switchState(Status::RUN);
	}
}

void AstriMaProcessor::cleanup() {
	// STOPPING
	auto startTime = std::chrono::steady_clock::now();
	printLog(std::cout, "cleaning up...");

	while (true) { // break at the timeout or when there are no more packets to process
		if (!receiver->isConnectedToClient() || !provider->isOpen() || receiveAndProcessPacket() < 1) {
			printLog(std::cout, "Stop processing");
			break;
		}
		auto currentTime = std::chrono::steady_clock::now();
		auto elapsedTime = std::chrono::duration_cast<std::chrono::seconds>(
				currentTime - startTime).count();
		if (elapsedTime >= closingTimeout) {
			printLog(std::cout,
					"stop processing but there were still other packets to process");
			break;
		}
	}
	receiver->closeConnectionToClient();
	provider->close();
}

void AstriMaProcessor::populatePacketStructuresMap() {
	for (std::pair<std::string, bool> pair : packetsFlags) { // for each packet type to process
		if (pair.second) { // if set to true, it searches for the corresponding .json file
			std::string filePath = std::string(ASTRI_PACKETS_PATH).append(
					"/AstriMA/").append(pair.first).append(".json");
			std::ifstream file(filePath);
			if (file.good()) {
				packetStructuresMap[pair.first] =
						new Packets::PacketStructureJson(filePath);
				file.close();
			} else {
				// File .json does not exists
				packetsFlags[pair.first] = false;
				printLog(std::cout,
						"File " + filePath + " does not exist, skipping "
								+ pair.first + " processing");
			}
		}
	}
}

std::string AstriMaProcessor::addPrefixToKey(std::string packetKey) {
	if (packetKey.compare("11") == 0 || packetKey.compare("14") == 0) {
		return packetKey.insert(0, "C");
	} else if (packetKey.compare("22") == 0) {
		return packetKey.insert(0, "S");
	} else if (packetKey.compare("101") == 0) {
		return packetKey.insert(0, "HK");
	} else if (packetKey.compare("102") == 0) {
		return packetKey.insert(0, "VAR");
	} else if (packetKey.compare("103") == 0) {
		return packetKey.insert(0, "VAR");
	} else if (packetKey.compare("151") == 0) {
		return packetKey.insert(0, "CMD");
	}
	return packetKey;
}

std::string AstriMaProcessor::getStateStr(Status state) {
	switch (state) {
	case Status::INIT:
		return std::string("INIT");
	case Status::STOP:
		return std::string("STOP");
	case Status::IDLE:
		return std::string("IDLE");
	case Status::RUN:
		return std::string("RUN");
	case Status::READY:
		return std::string("READY");
	default:
		return std::string("NONE");
	}
}
