#include <Astri_MA_Processor.h>

#include <Astri_MA_Generic.h>
#include <Astri_Horn_Generic.h>
#include <Redis_Receiver.h>
#include <KafkaAvro_Provider.h>
#include <File_Receiver.h>
#include <sstream>

using namespace inaf::oasbo::Processors;

AstriMaProcBuilder::AstriMaProcBuilder() {
	this->reset();
}
AstriMaProcBuilder::~AstriMaProcBuilder() {
	delete proc;
}
void AstriMaProcBuilder::reset() {
	this->proc = new AstriMaProcessor();
}

AstriMaProcBuilder* AstriMaProcBuilder::configFrom(
		Configurators::BaseConfigurator &conf) {
	proc->configurations.push_back(&conf);
	std::string target = std::string(this->config_target);
	conf.readConfigFromSource(target);
	std::map<std::string, std::string> params = conf.getConfig();

// ---- RECEIVER --------------------------------------------------------------------------
	std::string key = target + "_" + receiver_key;
	if (params.count(key) > 0) {
		if (params[key].compare(redis_rcv_key) == 0) {
			proc->receiver =
					Receivers::RedisReceiverBuilder().configFrom(conf)->getReceiver();
		}

		else if (params[key].compare(file_rcv_key) == 0) {
			proc->receiver =
					Receivers::FileReceiverBuilder().configFrom(conf)->getReceiver();
		} else { // not recognized, default to redis
			proc->receiver = Receivers::RedisReceiverBuilder().getReceiver();
		}
	}
// ---- PROVIDER --------------------------------------------------------------------
	key = target + "_" + provider_key;
	if (params.count(key) > 0) {
		if (params[key].compare(kafkaavro_prov_key) == 0) {
			proc->provider = Providers::KafkaAvroProviderBuilder().configFrom(
					conf)->getProvider();
		} else { // not recognized, default to KafkaAvro
			proc->provider = Providers::KafkaAvroProviderBuilder().configFrom(
					conf)->getProvider();
		}
	}

// ---- PACKET --------------------------------------------------------------------------
	key = target + "_" + packet_key;
	if (params.count(key) > 0) {
		if (params[key].compare(astri_ma_packet_key) == 0) {
			std::string ma_packet(
					std::string(ASTRI_PACKETS_PATH).append(
							"/AstriMA/AstriMAGeneric.json"));
			Packets::PacketStructureJson structure(ma_packet);
			proc->packet = new Packets::AstriMaGeneric(structure);
		} else if (params[key].compare(astri_horn_packet_key) == 0) {
			std::string ma_packet(
					std::string(ASTRI_PACKETS_PATH).append(
							"/AstriHorn/AstriHornGeneric.json"));
			Packets::PacketStructureJson structure(ma_packet);
			proc->packet = new Packets::AstriHornGeneric(structure);
		}

		else { // not recognized, default to astri_ma
			std::string ma_packet(
					std::string(ASTRI_PACKETS_PATH).append(
							"/AstriMA/AstriMAGeneric.json"));
			Packets::PacketStructureJson structure(ma_packet);
			proc->packet = new Packets::AstriMaGeneric(structure);
		}
	}
// ---- Packets to process --------------------------------------------
	key = target + "_" + typesubtype_key;
	if (params.count(key) > 0) {

		// packets value is expected in the form XX_XX_XX_XX...
		std::stringstream ss(params[key]);
		std::string item;
		while (std::getline(ss, item, '_')) {
			proc->packetsFlags[item] = true;
		}
	}

// ----- timeouts -----------------------------------------------------
	key = target + "_" + closingTimeout_key;
	if (params.count(key) > 0) {
		proc->closingTimeout = std::stoi(params[key]);
	}
	key = target + "_" + connectionTimeout_key;
	if (params.count(key) > 0) {
		proc->connectionTimeout = std::stoi(params[key]);
	}

	return this;
}

AstriMaProcBuilder* AstriMaProcBuilder::setReceiver(
		Receivers::BaseReceiver *value) {
	this->proc->receiver = value;
	return this;
}
AstriMaProcBuilder* AstriMaProcBuilder::setProvider(
		Providers::BaseProvider *value) {
	this->proc->provider = value;
	return this;
}

AstriMaProcBuilder* AstriMaProcBuilder::addPacketToProcess(std::string packet) {
	this->proc->packetsFlags[packet] = true;
	return this;
}

AstriMaProcBuilder* AstriMaProcBuilder::removePacketToProcess(
		std::string packet) {
	this->proc->packetsFlags[packet] = false;
	return this;
}

AstriMaProcessor* AstriMaProcBuilder::getProcessor() {
	AstriMaProcessor *result = this->proc;
	this->reset();
	return result;
}

const std::string AstriMaProcBuilder::config_target = "astriprocessor";

const std::string AstriMaProcBuilder::closingTimeout_key = "closingTimeout";
const std::string AstriMaProcBuilder::connectionTimeout_key = "connectionTimeout";

const std::string AstriMaProcBuilder::receiver_key = "receiver";
const std::string AstriMaProcBuilder::provider_key = "provider";
const std::string AstriMaProcBuilder::packet_key = "packet";
const std::string AstriMaProcBuilder::typesubtype_key = "toprocess";

const std::string AstriMaProcBuilder::redis_rcv_key = "redis";
const std::string AstriMaProcBuilder::file_rcv_key = "file";
const std::string AstriMaProcBuilder::kafkaavro_prov_key = "kafkaavro";
const std::string AstriMaProcBuilder::astri_ma_packet_key = "astrima";
const std::string AstriMaProcBuilder::astri_horn_packet_key = "astrihorn";

