/**
 * @file main.cpp
 * @brief This file contains the main function for the Astri MA Processor.
 * 
 * The Astri MA Processor is responsible for processing data using the AstriMaProcBuilder.
 * It reads configuration from YAML, MySQL, and command line arguments.
 * The program starts and stops the processor.
 * 
 * The main function sets up signal handlers for CTRL-C and CTRL-TERM.
 * It also prints the program start time and configuration parameters.
 * The program runs until the processor's current state is STOP.
 * Finally, it cleans up resources and prints the termination message.
 */
#include <Astri_MA_Processor.h>
#include <Astri_MA_Generic.h>

#include <Yaml_Conf.h>
#include <CL_Conf.h>
#include <Mysql_Conf.h>

#include <csignal>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <atomic>
#include <thread>
#include <ctime>
#include <iomanip>

#ifndef ASTRIPROC_CONFIG_PATH
#define ASTRIPROC_CONFIG_PATH std::string(std::getenv("HOME")).append("/BIAS/config/AstriProcs")
#endif

using namespace inaf::oasbo;
using namespace inaf::oasbo::Processors;

AstriMaProcessor *proc;

void my_handler(int s) {
	time_t now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[Astri MA Processor]\t" << "Termination" << std::endl;
	proc->stop(5);
}

int main(int argc, char **argv) {
///////// PROCS CONF -------------------------------------------------------------------------------------------------
	Configurators::YamlConfigurator procsyamlconf(
			std::string(ASTRIPROC_CONFIG_PATH).append("/config.yaml"));
	Configurators::MysqlConfigurator mysqlConf("127.0.0.1", 3306, "bias",
			"BIAS1ast01$1", "BIAS");
	Configurators::CLConfigurator clConf(argc, argv);
	proc =
			Processors::AstriMaProcBuilder().configFrom(procsyamlconf)->configFrom(
					mysqlConf)->configFrom(clConf)->getProcessor();

	std::map<std::string, std::string> procsConf;
	for (auto [k, v] : procsyamlconf.getConfig())
		procsConf[k] = v;
	for (auto [k, v] : mysqlConf.getConfig())
		procsConf[k] = v;
	for (auto [k, v] : clConf.getConfig())
		procsConf[k] = v;

	time_t now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[Astri MA Processor]\t" "Starting program with params: "
			<< std::endl;
	for (auto [k, v] : procsConf)
		std::cout << k << " : " << v << std::endl;
	std::cout << "-------------------------------" << std::endl;

	// CTRL-C handler
	std::signal(SIGINT, [](int signal) {
		my_handler(signal);
	});
	std::signal(SIGTERM, [](int signal) {
		my_handler(signal);
	});

	// START
	proc->start();

	while (proc->getCurrentState() != AstriMaProcessor::Status::STOP) {
		std::this_thread::sleep_for(std::chrono::seconds(1));
	}

	delete proc;

	now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[Astri MA Processor]\t" << "Terminated normally"
			<< std::endl;
	return EXIT_SUCCESS;
}
